{
              CsarChiffre  September 2005 Michael Stapelberg

  Beschreibung: Das Programm verschlsselt Texte mithilfe der Csar-Methode.
                Dabei stehen viele Optionen zur Verfgung: Welche Teile
                des Alphabets verwendet werden sollen, ob das chiffrierte
                Alphabet nur durch Verschiebung oder durch ein Codewort
                erstellt werden soll und zu guter letzt die "Starke
                Verschlsselung", ein eigenes Verfahren, welches
                Hufigkeitsanalysen deutlich erschweren soll.

  Lizenz:       Open-Source, solange der Quelltext oder Teile davon nicht
                fr kommerzielle Zwecke verwendet wird.

  Bemerkungen:  Die Komponenten wurden im Stil <TypAbkrzung><Name> benannt,
                zum Beispiel "CBDechiffrieren" fr die Checkbox, die bestimmt,
                ob man momentan dechiffriert oder nicht. Dadurch sortiert Delphi
                im Fenster "Objekt-Hierarchie" die Komponenten schner und
                man erkennt auch im Code sofort, welchen Typ eine Komponente
                hat.

                Es werden nur Standardkomponenten verwendet, die Komponente
                SpinEdit findet man im Tab "Beispiele".
}










unit MainForm;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Spin, ExtCtrls;

type
  TAlphabet = record
    OriginalAlphabet:string;
    CryptedAlphabet:string;
  end;

  TForm1 = class(TForm)
    MEingangstext: TMemo;
    MAusgangsText: TMemo;
    Label1: TLabel;
    Label2: TLabel;
    AlphabetNormal: TLabel;
    AlphabetChiffriert: TLabel;
    Schluessel: TSpinEdit;
    GBAlphabetUndSchluessel: TGroupBox;
    GBEingangstext: TGroupBox;
    GBAusgangsText: TGroupBox;
    POptionenUndKontrolle: TPanel;
    RBAlphabetVerschieben: TRadioButton;
    Label3: TLabel;
    RBCodewortVerwenden: TRadioButton;
    ECodeWort: TEdit;
    Label4: TLabel;
    CBStarkeVerschluesselung: TCheckBox;
    Label5: TLabel;
    CBDechiffrieren: TCheckBox;
    CBSonderzeichenBehandeln: TCheckBox;
    CBKleinbuchstabenBehandeln: TCheckBox;
    GroupBox4: TGroupBox;
    CBZahlenBehandeln: TCheckBox;
    CBAutomatischZwischenablage: TCheckBox;
    GBOptionenUndKontrolle: TGroupBox;
    Label6: TLabel;
    CBLeerzeichenEntfernen: TCheckBox;
    CBGrossKleinVermischen: TCheckBox;
    procedure StartButtonClick(Sender: TObject);
    procedure SchluesselAllgChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
  private
    Alphabet:TAlphabet;
    procedure AlphabetErneuern;
    function CaesarChiffre(const AText:string; AAlphabet:TAlphabet):string;
  end;

var
  Form1: TForm1;

implementation

{$R *.dfm}

// -------------------------------------------------------------------------
// Allgemeine Funktionen, die im Laufe des Programms verwendet werden
// -------------------------------------------------------------------------

// Inline IF:
// Funktioniert wie ein normales "if", kann allerdings als Funktion verwendet
// werden, siehe das (<condition>?<true>:<else>)-Konstrukt in C.
function iif(const ACondition:boolean; const ATrue,AFalse:integer):integer;
begin
if (ACondition) then
  result:=ATrue
else result:=AFalse;
end;

function iifString(const ACondition:boolean; const ATrue,AFalse:string):string;
begin
if (ACondition) then
  result:=ATrue
else result:=AFalse;
end;

// -------------------------------------------------------------------------
// Programmspezifische Funktionen und Prozeduren
// -------------------------------------------------------------------------

// Diese Prozedur erneuert das Original- und das Chiffrierte Alphabet.
// Dadurch werden die meisten Optionen sowie der Algorithmus umgesetzt
// Die Prozedur wird jedes mal aufgerufen, wenn eine der Schlsseloptionen
// einen anderen Wert erlangt

function Scramble(const AInput:string):string;
var Text:string;
    c:integer;
begin
Text:=AInput;
for c:=1 to length(Text) do
  if (Random(2) = 0) then
    Text[c]:=Uppercase(Text[c])[1]
  else Text[c]:=Lowercase(Text[c])[1];

result:=Text;
end;

procedure TForm1.AlphabetErneuern;
var c,Position,NearestPos,GeaenderteBuchstaben:integer;
    TempAlphabet,TempAlphabetOriginal:string;
function GetNearestCharPos(const ABaseChar:integer):integer;
var LowestValue,x:integer;
begin
result:=0;
LowestValue:=999; // not the finest way but should do it :|
for x:=1 to length(TempAlphabet) do
  if  (Ord(TempAlphabet[x]) > ABaseChar) and
      (Ord(TempAlphabet[x]) < LowestValue) then begin
    result:=x;
    LowestValue:=Ord(TempAlphabet[x]);
  end;
end;
begin
Alphabet.OriginalAlphabet:='';
Alphabet.CryptedAlphabet:='';

// Alle Werte knnen in einer ASCII-Tabelle nachgelesen werden
// (zum Beispiel auf http://www.asciitable.com)

// Grobuchstaben A-Z
for c:=65 to 90 do
  Alphabet.OriginalAlphabet:=Alphabet.OriginalAlphabet+Chr(c);

// Kleinbuchstaben a-z
if (CBKleinbuchstabenBehandeln.Checked) then
  for c:=97 to 122 do
    Alphabet.OriginalAlphabet:=Alphabet.OriginalAlphabet+Chr(c);

if (CBSonderzeichenBehandeln.Checked) then begin
  // Sonderzeichen (teil1)
  for c:=33 to 47 do
    Alphabet.OriginalAlphabet:=Alphabet.OriginalAlphabet+Chr(c);
  // Sonderzeichen (teil2)
  for c:=58 to 64 do
    Alphabet.OriginalAlphabet:=Alphabet.OriginalAlphabet+Chr(c);
end;

// Zahlen 0-9
if (CBZahlenBehandeln.Checked) then
  for c:=48 to 57 do
    Alphabet.OriginalAlphabet:=Alphabet.OriginalAlphabet+Chr(c);

Schluessel.MaxValue:=length(Alphabet.OriginalAlphabet);

if (RBAlphabetVerschieben.Checked) then begin
  // Methode: Alphabet verschieben
  TempAlphabet:=Alphabet.OriginalAlphabet+Alphabet.OriginalAlphabet;
  for c:=Schluessel.Value to Schluessel.MaxValue+Schluessel.Value-1 do
    Alphabet.CryptedAlphabet:=Alphabet.CryptedAlphabet+copy(TempAlphabet,c+1,1);
end else if (RBCodewortVerwenden.Checked) then begin
  TempAlphabet:=Alphabet.OriginalAlphabet;
  for c:=1 to length(ECodeWort.Text) do begin
    if (Pos(copy(ECodeWort.Text,c,1),Alphabet.CryptedAlphabet) > 0) then
      continue;

    Position:=Pos(copy(ECodeWort.Text,c,1),TempAlphabet);

    // TODO: Maybe error check?
    if (Position = 0) then
      continue;

    Alphabet.CryptedAlphabet:=Alphabet.CryptedAlphabet+copy(TempAlphabet,Position,1);
    Delete(TempAlphabet,Position,1);
  end;
  GeaenderteBuchstaben:=length(Alphabet.CryptedAlphabet);
  for c:=1 to length(Alphabet.CryptedAlphabet) do
    if (Alphabet.CryptedAlphabet[c] <> Alphabet.OriginalAlphabet[c]) then
      break
    else Dec(GeaenderteBuchstaben);
  Label5.Caption:='Hinweis: Ihr Codewort hat effektiv '+IntToStr(GeaenderteBuchstaben)+' Buchstaben gendert.';

  if (length(Alphabet.CryptedAlphabet) > 0) then
    NearestPos:=GetNearestCharPos(Ord(Alphabet.CryptedAlphabet[length(Alphabet.CryptedAlphabet)]))
  else NearestPos:=1;
  TempAlphabetOriginal:=TempAlphabet;
  TempAlphabet:=TempAlphabet+TempAlphabet;
  for c:=0 to Length(TempAlphabetOriginal)-1 do
    Alphabet.CryptedAlphabet:=Alphabet.CryptedAlphabet+copy(TempAlphabet,c+NearestPos,1);

end;

if (CBDechiffrieren.Checked) then begin
  AlphabetNormal.Caption:=Alphabet.CryptedAlphabet;
  AlphabetChiffriert.Caption:=Alphabet.OriginalAlphabet;
  // Werte mssen auch wieder zurckbertragen werden, da wir mit Alphabet
  // weiterarbeiten spter
  Alphabet.OriginalAlphabet:=AlphabetNormal.Caption;
  if (not CBStarkeVerschluesselung.Checked) then
    Alphabet.CryptedAlphabet:=AlphabetChiffriert.Caption;
end else begin
  AlphabetNormal.Caption:=Alphabet.OriginalAlphabet;
  if (not CBStarkeVerschluesselung.Checked) then
    AlphabetChiffriert.Caption:=Alphabet.CryptedAlphabet;
end;

if (RBCodewortVerwenden.Checked and CBStarkeVerschluesselung.Checked) then
  AlphabetChiffriert.Caption:='Es kann kein chiffriertes Alphabet fr diese Methode angezeigt werden';

Label6.Visible:=(length(Alphabet.CryptedAlphabet) < length(MEingangsText.Text));
CBStarkeVerschluesselung.Enabled:=RBCodewortVerwenden.Checked;
end;

// -------------------------------------------------------------------------

// Die eigentliche Verschlsselungsmethode
// Hinweis: Kann in ihrer jetztigen Form nicht in anderen Projekten
// verwendet werden, da sie auf Eigenschaften der Form1 zugreift (Optionen)

function TForm1.CaesarChiffre(const AText:string; AAlphabet:TAlphabet):string;
var BuchstabenZaehler,AlphabetLaenge:integer;
    Text, CryptAlphabet:string;
begin
  Text:=AText;

  // Wenn die Kleinbuchstaben nicht behandelt werden wandeln wir den Text
  // vorher in Grobuchstaben um
  if (not Form1.CBKleinbuchstabenBehandeln.Checked) then
    Text:=Uppercase(Text)
  else
    if (Form1.CBGrossKleinVermischen.Checked) then
      Text:=Scramble(Text);
      
  result:='';

  // Wir kopieren das verschlsselte Alphabet in eine Variable, damit wir
  CryptAlphabet:=AAlphabet.CryptedAlphabet;
  AlphabetLaenge:=length(CryptAlphabet);
  // es solange duplizieren knnen, bis es mindestens so lange ist, wie
  // der Klartext + das Alphabet (fr die "Starke Verschlsselung" Vorraussetzung)
  while (length(CryptAlphabet) < length(AText)+AlphabetLaenge) do
    CryptAlphabet:=CryptAlphabet+CryptAlphabet;

  for BuchstabenZaehler:=1 to length(Text) do begin
    // Sollte das Zeichen nicht in unserem OriginalAlphabet vorhanden sein
    // bertragen wir es einfach ins Ergebnis
    if (Pos(Text[BuchstabenZaehler],AAlphabet.OriginalAlphabet) = 0) then
      result:=result+Text[BuchstabenZaehler]

    // Ansonsten (Zeichen ist vorhanden):
    // Das Ergebnis(result) wird erweitert mit:
    else result:=result+
                  // dem Zeichen aus CryptAlphabet
                  copy( CryptAlphabet,
                        // das die Position des Zeichens im OriginalAlphabet hat
                        Pos(Text[BuchstabenZaehler],AAlphabet.OriginalAlphabet)+
                        // addiert mit der momentanen Position in Klartext AText
                        // wenn Form1.CBStarkeVerschluesselung.Checked ist
                        iif(Form1.RBCodewortVerwenden.Checked and Form1.CBStarkeVerschluesselung.Checked,
                            // Wenn dechiffriert wird
                            iif(Form1.CBDechiffrieren.Checked,
                                // Wird das Vorzeichen umgedreht
                                (BuchstabenZaehler-1) * -1,
                                // Wenn nicht dechiffriert wird, positiven Wert zurckgeben
                                BuchstabenZaehler-1),
                            // Wenn nicht "stark" verschlsselt wird, nichts addieren
                            0)
                        // Lnge des zu kopierenden Bereichs: 1 Zeichen
                        ,1);
  end;
end;

// -------------------------------------------------------------------------
// Funktionen fr die grafische Oberflche (=GUI)
// -------------------------------------------------------------------------

procedure TForm1.StartButtonClick(Sender: TObject);
var Zaehler:integer;
begin
MAusgangsText.Clear;
for Zaehler:=0 to (MEingangstext.Lines.Count-1) do
  // Nach verbesserter Methode chiffrieren:
  MAusgangsText.Lines.Add(CaesarChiffre(MEingangstext.Lines.Strings[Zaehler],Alphabet));

if (CBAutomatischZwischenablage.Checked) then begin
  MAusgangsText.SelectAll;
  MAusgangsText.CopyToClipboard;
end;

end;

procedure TForm1.SchluesselAllgChange(Sender: TObject);
var Zaehler:integer;
begin
// Der Wert eines SpinEdits kann, wenn er manuell eingegeben wird,
// von den eingestellten MinValue- und MaxValue-Werten abweichen.
// Wir korrigieren das hier.
if (Schluessel.Value < Schluessel.MinValue) then
  Schluessel.Value:=Schluessel.MinValue
else if (Schluessel.Value > Schluessel.MaxValue) then
  Schluessel.Value:=Schluessel.MaxValue;
  
AlphabetErneuern;
MAusgangsText.Clear;
for Zaehler:=0 to (MEingangstext.Lines.Count-1) do
  // Nach verbesserter Methode chiffrieren:
  MAusgangsText.Lines.Add(StringReplace(CaesarChiffre(MEingangstext.Lines.Strings[Zaehler],Alphabet),' ',iifString(CBLeerzeichenEntfernen.Checked,'',' '),[rfReplaceAll]));

if (CBAutomatischZwischenablage.Checked) then begin
  MAusgangsText.SelectAll;
  MAusgangsText.CopyToClipboard;
end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin

// Zufallsgenerator initialisieren
Randomize;

// DoubleBuffered zeichnet zuerst in den Speicher, statt direkt auf den
// Bildschirm. Dadurch wird ein nerviges Flackern vermieden, wenn Windows
// unsere Form oft hintereinander neu zeichnen lsst.

DoubleBuffered:=true;
MEingangstext.DoubleBuffered:=true;
MAusgangsText.DoubleBuffered:=true;
GBEingangsText.DoubleBuffered:=true;
GBAusgangsText.DoubleBuffered:=true;
GBAlphabetUndSchluessel.DoubleBuffered:=true;
GBOptionenUndKontrolle.DoubleBuffered:=true;

// Chiffrierte Darstellung intialisieren
SchluesselAllgChange(nil);
end;

end.
